import asyncio
import time

from .. import fhem
from ..xiaomi_gateway3 import xiaomi_gateway3
from .. import fhem_pythonbinding as fhepy


class xiaomi_gateway3_device:
    def __init__(self, logger):
        self.logger = logger
        self.gateway = None
        self.device_details = None
        self.last_update = 0
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        self.hash = hash

        if len(args) < 5:
            return (
                "Usage: define devname PythonModule xiaomi_gateway3_device <IP> <DID>"
            )

        self.gw_name = args[3]
        self.did = args[4]

        hash["GATEWAY"] = self.gw_name
        hash["DID"] = self.did

        await fhem.readingsSingleUpdateIfChanged(self.hash, "presence", "offline", 1)

        asyncio.create_task(self.connect_gw())
        self.offline_check_task = asyncio.create_task(self.offline_check())

        return ""

    async def offline_check(self):
        while True:
            if time.time() - self.last_update > 3700:
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "presence", "offline", 1
                )
            await asyncio.sleep(300)

    async def connect_gw(self):
        while self.gateway is None:
            self.gateway = fhepy.getFhemPyDeviceByName(self.gw_name)
            if self.gateway:
                try:
                    self.gateway.register_device(self.did, self)
                    await fhem.readingsSingleUpdateIfChanged(
                        self.hash, "presence", "online", 1
                    )
                except:
                    self.gateway = None
                    pass
            else:
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "state", f"gateway {self.gw_name} not found", 1
                )
            await asyncio.sleep(10)

    # FHEM FUNCTION
    async def Undefine(self, hash):
        self.offline_check_task.cancel()
        return

    # FHEM FUNCTION
    async def Set(self, hash, args, argsh):
        return

    async def update(self, data):
        self.last_update = time.time()

        attr_conf = {
            "lumi.sensor_magnet.v2": {
                "devStateIcon": "open:fts_door_open\@red close:fts_door\@green",
                "icon": "tuer_fenster_kontakt",
            },
            "lumi.sensor_ht.v1": {
                "stateFormat": "temperature °C, humidity %",
                "icon": "temp_temperature",
            },
            "lumi.sensor_ht.v2": {
                "stateFormat": "temperature °C, humidity %, pressure kPa",
                "icon": "temp_temperature",
            },
            "lumi.sensor_motion.v1": {
                "devStateIcon": "motion:motion_detector\@red off:motion_detector\@green no_motion:motion_detector\@green",
                "icon": "people_sensor",
            },
            "lumi.gateway.mgl03": {
                "stateFormat": "presence",
                "devStateIcon": "online:it_wifi\@red offline:it_wifi\@red",
                "icon": "tradfri_gateway",
            },
        }

        # first update, set attributes and device readings like model, sid, ...
        if self.device_details is None:
            self.device_details = self.gateway.get_device(self.did)
            if self.device_details["model"] in attr_conf:
                for attr in attr_conf[self.device_details["model"]]:
                    if await fhem.AttrVal(self.hash["NAME"], attr, "") == "":
                        await fhem.CommandAttr(
                            self.hash,
                            f"{self.hash['NAME']} {attr} {attr_conf[self.device_details['model']][attr]}",
                        )
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "model", self.device_details["model"], 1
            )
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "sid", self.device_details["sid"], 1
            )
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "mac", self.device_details["mac"], 1
            )
            if "zb_ver" in self.device_details:
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "zb_ver", self.device_details["zb_ver"], 1
                )

        # device is online
        await fhem.readingsSingleUpdateIfChanged(self.hash, "presence", "online", 1)

        # update data
        for reading in data:
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, reading, data[reading], 1
            )



# THANKS to https://github.com/AlexxIT/XiaomiGateway3 for the HASS implementation
# which I used as a base for the FHEM implementation

import base64
import json
import asyncio
import socket
import re
import functools
from telnetlib import Telnet
import time
from asyncio_mqtt import Client, MqttError

from .miio_fix import Device
from .unqlite import Unqlite

from .utils import GLOBAL_PROP
from . import utils

from .. import fhem
from .. import utils as fpyutils

RE_NWK_KEY = re.compile(r"lumi send-nwk-key (0x.+?) {(.+?)}")


class xiaomi_gateway3:
    def __init__(self, logger):
        self.logger = logger
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        self.hash = hash

        if len(args) < 5:
            return "Usage: define devname PythonModule xiaomi_gateway3 <IP> <token>"

        if await fhem.AttrVal(self.hash["NAME"], "icon", "") == "":
            await fhem.CommandAttr(self.hash, self.hash["NAME"] + " icon mqtt")
        await fhem.readingsSingleUpdateIfChanged(hash, "state", "disconnected", 1)

        hash["HOST"] = args[3]
        hash["TOKEN"] = args[4]

        self.host = args[3]
        self.token = args[4]

        asyncio.create_task(self.connect_gw())

        return ""

    # FHEM FUNCTION
    async def Undefine(self, hash):
        return

    # FHEM FUNCTION
    async def Set(self, hash, args, argsh):
        return ""

    def register_device(self, did, upd_listener):
        self.gw.register_device(did, upd_listener)

    def get_device(self, did):
        return self.gw.get_device(did)

    async def connect_gw(self):
        self.gw = Gateway(self.logger, self.hash, self.host, self.token)
        # connect to gateway
        await self.gw.connect()
        # create task which handles MQTT messages
        asyncio.create_task(self.gw.connect_mqtt())


class Gateway:
    def __init__(self, logger, hash, host, token):
        self.hash = hash
        self.logger = logger
        self.host = host
        self.token = token
        self.devices = {}
        self.miio = Device(host, token)
        self.child_devices = {}
        self.connected = False
        self.pair_model = None
        self.pair_payload = None

    def register_device(self, did, upd_listener):
        if did not in self.child_devices:
            self.child_devices[did] = []
        self.child_devices[did].append(upd_listener)
        # report initial states
        asyncio.create_task(self.report(did))

    def get_device(self, did):
        if did in self.devices:
            return self.devices[did]
        return None

    async def connect(self):
        # first connect
        await fpyutils.run_blocking(functools.partial(self.thread_blocking_connect))
        await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "connected", 1)
        await self.create_devices()
        await self.report_all()
        # start check_connection task to reconnect on gw reboot
        asyncio.create_task(self.check_connection())

    async def check_connection(self):
        while True:
            if self.connected is False:
                await fpyutils.run_blocking(
                    functools.partial(self.thread_blocking_connect)
                )
                if self.connected:
                    await self.create_devices()
                    await self.report_all()
            # sleep 30s
            await asyncio.sleep(30)

    def thread_blocking_connect(self, keepconnection=0):
        while self.connected is False:
            if self._miio_connect():
                self.update_devices()
            else:
                time.sleep(30)
        # connected

    def update_devices(self):
        devices = self._get_devices_v3()
        if devices is None:
            self._enable_telnet()
            self._enable_mqtt()
        else:
            self.setup_devices(devices)

    def setup_devices(self, devices: list):
        """Add devices to hass."""
        for device in devices:
            desc = utils.get_device(device["model"])
            if not desc:
                self.logger.debug(f"Unsupported model: {device}")
                continue

            self.logger.debug(f"{self.host} | Setup device {device['model']}")

            device.update(desc)

            # update params from config
            default_config = self.devices.get(device["mac"])
            if default_config:
                device.update(default_config)

            self.devices[device["did"]] = device

    async def create_devices(self):
        for did in self.devices:
            if not await fhem.checkIfDeviceExists(
                self.hash, "PYTHONTYPE", "xiaomi_gateway3_device", "DID", did
            ):
                devname = (
                    "".join(filter(str.isalnum, self.devices[did]["model"]))
                    + "_"
                    + self.devices[did]["sid"]
                )
                await fhem.CommandDefine(
                    self.hash,
                    devname
                    + " PythonModule xiaomi_gateway3_device "
                    + self.hash["NAME"]
                    + " "
                    + did,
                )
                # avoid fhem commands overload
                await asyncio.sleep(3)
            # avoid fhem commands overload
            await asyncio.sleep(1)

    async def report(self, did):
        if did in self.devices:
            device = self.devices[did]
            if "init" in device:
                if did in self.child_devices:
                    for listener in self.child_devices[did]:
                        await listener.update(device["init"])

    async def report_all(self):
        for did in self.devices:
            await self.report(did)

    def _miio_connect(self) -> bool:
        try:
            self.miio.send_handshake()
            return True
        except:
            self.logger.debug(f"{self.host} | Can't send handshake")
            return False

    def _enable_telnet(self):
        self.logger.debug(f"{self.host} | Try enable telnet")
        try:
            resp = self.miio.send("enable_telnet_service")
            return resp[0] == "ok"
        except Exception as e:
            self.logger.exception(f"Can't enable telnet: {e}")
            return False

    def _enable_mqtt(self):
        self.logger.debug(f"{self.host} | Try run public MQTT")
        try:
            telnet = Telnet(self.host)
            telnet.read_until(b"login: ")
            telnet.write(b"admin\r\n")
            telnet.read_until(b"\r\n# ")  # skip greeting

            # enable public mqtt
            telnet.write(b"killall mosquitto\r\n")
            telnet.read_until(b"\r\n")  # skip command
            time.sleep(0.5)  # it's important to wait
            telnet.write(b"mosquitto -d\r\n")
            telnet.read_until(b"\r\n")  # skip command
            time.sleep(0.5)  # it's important to wait

            # fix CPU 90% full time bug
            telnet.write(b"killall zigbee_gw\r\n")
            telnet.read_until(b"\r\n")  # skip command
            time.sleep(0.5)  # it's important to wait

            telnet.close()
            return True
        except Exception as e:
            self.logger.debug(f"Can't run MQTT: {e}")
            return False

    async def connect_mqtt(self):
        try:
            async with Client(self.host) as client:
                self.mqtt = client
                async with client.filtered_messages("#") as messages:
                    await client.subscribe("#")
                    async for message in messages:
                        self.logger.debug(
                            f"Message from MQTT: {message.topic}: {message.payload.decode()}"
                        )
                        try:
                            if message.topic == "zigbee/send":
                                payload = json.loads(message.payload.decode())
                                # convert to miio msg
                                msg = self.process_message(payload)
                                if msg["did"] in self.child_devices:
                                    for listener in self.child_devices[msg["did"]]:
                                        await listener.update(msg["data"])
                            elif message.topic.endswith("/heartbeat"):
                                for listener in self.child_devices["lumi.0"]:
                                    await listener.update(
                                        json.loads(message.payload.decode())
                                    )
                            elif self.pair_model and message.topic.endswith(
                                "/commands"
                            ):
                                self.process_pair(message.payload)
                        except:
                            self.logger.exception("Failed to handle MQTT message")
        except:
            self.connected = False
            self.logger.exception("Failed to connect to MQTT server: " + self.host)

    def process_pair(self, raw: bytes):
        # get shortID and eui64 of paired device
        if b"lumi send-nwk-key" in raw:
            # create model response
            payload = (
                f"0x18010105000042{len(self.pair_model):02x}"
                f"{self.pair_model.encode().hex()}"
            )
            m = RE_NWK_KEY.search(raw.decode())
            self.pair_payload = json.dumps(
                {
                    "sourceAddress": m[1],
                    "eui64": "0x" + m[2],
                    "profileId": "0x0104",
                    "clusterId": "0x0000",
                    "sourceEndpoint": "0x01",
                    "destinationEndpoint": "0x01",
                    "APSCounter": "0x01",
                    "APSPlayload": payload,
                },
                separators=(",", ":"),
            )

        # send model response "from device"
        elif b"zdo active " in raw:
            mac = ""  # self.device['mac'][2:].upper()
            self.mqtt.publish(f"gw/{mac}/MessageReceived", self.pair_payload)

    def _get_devices_v3(self):
        """Load device list via Telnet."""
        self.logger.debug(f"{self.host} | Read devices")
        try:
            telnet = Telnet(self.host, timeout=5)
            telnet.read_until(b"login: ")
            telnet.write(b"admin\r\n")
            telnet.read_until(b"\r\n# ")  # skip greeting

            # https://github.com/AlexxIT/XiaomiGateway3/issues/14
            # fw 1.4.6_0012 and below have one zigbee_gw.db file
            # fw 1.4.6_0030 have many json files in this folder
            telnet.write(b"cat /data/zigbee_gw/* | base64\r\n")
            telnet.read_until(b"\r\n")  # skip command
            raw = telnet.read_until(b"# ")
            raw = base64.b64decode(raw)
            if raw.startswith(b"unqlite"):
                db = Unqlite(raw)
                data = db.read_all()
            else:
                raw = re.sub(br"}\s+{", b",", raw)
                data = json.loads(raw)

            devices = []

            # data = {} or data = {'dev_list': 'null'}
            dev_list = json.loads(data.get("dev_list", "null")) or []
            self.logger.debug(f"{self.host} | Load {len(dev_list)} zigbee devices")

            for did in dev_list:
                model = data[did + ".model"]
                desc = utils.get_device(model)

                # skip unknown model
                if desc is None:
                    self.logger.debug(f"Unsupported model: {model}")
                    continue

                retain = json.loads(data[did + ".prop"])["props"]

                params = retain
                # params = {
                #     p[2]: retain.get(p[1])
                #     for p in desc['params']
                #     if p[1] is not None
                # }

                # fix some param values
                for k, v in list(params.items()):
                    if k in ("temperature", "humidity", "pressure"):
                        params[k] = v / 100.0
                    elif k in ("voltage"):
                        params[k] = v / 1000.0
                    elif k in ("battery") and v:
                        params["batteryPercentage"] = params["battery"]
                        params["battery"] = (
                            "low" if params["batteryPercentage"] < 25 else "ok"
                        )
                    elif k in ("status"):
                        params["state"] = params["status"]
                        del params["status"]

                self.logger.debug(f"{self.host} | {model} retain: {params}")

                device = {
                    "did": did,
                    "sid": did.split(".")[1],
                    "mac": "0x" + data[did + ".mac"],
                    "model": data[did + ".model"],
                    "type": "zigbee",
                    "zb_ver": data[did + ".version"],
                    "init": params,
                }
                devices.append(device)

            telnet.write(b"cat /data/zigbee/coordinator.info\r\n")
            telnet.read_until(b"\r\n")  # skip command
            raw = telnet.read_until(b"# ")

            device = self.miio.info()
            devices.append(
                {
                    "did": "lumi.0",
                    "sid": "0",
                    "mac": device.mac_address,  # wifi mac!!!
                    "model": device.model,
                }
            )

            self.connected = True

            return devices

        except (ConnectionRefusedError, socket.timeout):
            return None

        except Exception as e:
            self.logger.debug(f"Can't read devices: {e}")
            return None

    def process_message(self, data: dict):
        cmd = data["cmd"]
        if data["cmd"] == "heartbeat":
            # don't know if only one item
            assert len(data["params"]) == 1, data

            data = data["params"][0]
            pkey = "res_list"
        elif data["cmd"] == "report":
            pkey = "params" if "params" in data else "mi_spec"
        elif data["cmd"] == "write_rsp":
            pkey = "results"
        else:
            raise Exception(f"Unsupported cmd: {data}")

        did = data["did"]

        device = self.devices[did]
        payload = {}

        # convert codes to names
        for param in data[pkey]:
            if param.get("error_code", 0) != 0:
                continue

            prop = (
                param["res_name"]
                if "res_name" in param
                else f"{param['siid']}.{param['piid']}"
            )

            if prop in GLOBAL_PROP:
                prop = GLOBAL_PROP[prop]
            else:
                prop = next((p[2] for p in device["params"] if p[0] == prop), prop)

            if prop in ("temperature", "humidity", "pressure"):
                payload[prop] = param["value"] / 100.0
            elif prop in ("voltage"):
                payload[prop] = param["value"] / 1000.0
            elif prop in ("battery"):
                # xiaomi light sensor
                payload["batteryPercentage"] = param["value"]
                payload["battery"] = (
                    "low" if payload["batteryPercentage"] < 25 else "ok"
                )
            elif prop in ("status"):
                payload["state"] = param["value"]
            elif prop == "contact":
                payload["state"] = "open" if param["value"] == 1 else "close"
            elif prop == "angle":
                # xiaomi cube 100 points = 360 degrees
                payload[prop] = param["value"] * 4
            elif prop == "duration":
                # xiaomi cube
                payload[prop] = param["value"] / 1000.0
            else:
                payload[prop] = param["value"]

        self.logger.debug(
            f"{self.host} | {device['did']} {device['model']} <= " f"{payload}"
        )

        if did in self.devices:
            self.devices[did]["init"] = payload
        else:
            # TODO create device
            pass

        return {
            "did": did,
            "sid": device["sid"],
            "model": device["model"],
            "cmd": cmd,
            "data": payload,
        }



import asyncio
import functools

from micloud import MiCloud

from .. import utils
from .. import fhem


class xiaomi_tokens:
    def __init__(self, logger):
        self.logger = logger
        self._username = None
        self._password = None
        self._country = "de"
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        self.hash = hash
        await fhem.readingsSingleUpdateIfChanged(hash, "state", "active", 1)
        return ""

    # FHEM FUNCTION
    async def Undefine(self, hash):
        return

    # FHEM FUNCTION
    async def Set(self, hash, args, argsh):
        set_list_conf = {
            "username": {"args": ["username"]},
            "password": {"args": ["password"]},
            "country": {"args": ["country"], "options": "de,cn,sg"},
            "get_tokens": {},
        }
        return await utils.handle_set(set_list_conf, self, hash, args, argsh)

    async def set_username(self, hash, params):
        self._username = params["username"]
        return ""

    async def set_password(self, hash, params):
        self._password = params["password"]
        return ""

    async def set_country(self, hash, params):
        self._country = params["country"]
        return ""

    async def set_get_tokens(self, hash):
        if self._username and self._password:
            asyncio.create_task(self.obtain_tokens())
        else:
            return "Please set username & password first!"

    async def obtain_tokens(self):
        await utils.run_blocking(functools.partial(self.thread_get_tokens))
        await fhem.readingsBeginUpdate(self.hash)
        for dev in self._device_list:
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, dev["did"] + "_name", dev["name"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, dev["did"] + "_token", dev["token"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, dev["did"] + "_model", dev["model"]
            )
        await fhem.readingsEndUpdate(self.hash, 1)

    def thread_get_tokens(self):
        mc = MiCloud(self._username, self._password)
        mc.login()
        self._device_list = mc.get_devices(country=self._country)



class Unqlite:
    page_size = 0
    pos = 0

    def __init__(self, raw: bytes):
        self.raw = raw
        self.read_db_header()

    @property
    def size(self):
        return len(self.raw)

    def read(self, length: int):
        self.pos += length
        return self.raw[self.pos - length : self.pos]

    def read_int(self, length: int):
        return int.from_bytes(self.read(length), "big")

    def read_db_header(self):
        assert self.read(7) == b"unqlite", "Wrong file signature"
        assert self.read(4) == b"\xDB\x7C\x27\x12", "Wrong DB magic"
        creation_time = self.read_int(4)
        sector_size = self.read_int(4)
        self.page_size = self.read_int(4)
        assert self.read(6) == b"\x00\x04hash", "Unsupported hash"

    # def read_header2(self):
    #     self.pos = self.page_size
    #     magic_numb = self.read(4)
    #     hash_func = self.read(4)
    #     free_pages = self.read_int(8)
    #     split_bucket = self.read_int(8)
    #     max_split_bucket = self.read_int(8)
    #     next_page = self.read_int(8)
    #     num_rect = self.read_int(4)
    #     for _ in range(num_rect):
    #         logic_page = self.read_int(8)
    #         real_page = self.read_int(8)

    def read_cell(self):
        key_hash = self.read(4)
        key_len = self.read_int(4)
        data_len = self.read_int(8)
        next_offset = self.read_int(2)
        overflow_page = self.read_int(8)
        if overflow_page:
            self.pos = overflow_page * 0x1000 + 8
            data_page = self.read_int(8)
            data_offset = self.read_int(2)
            name = self.read(key_len)
            self.pos = data_page * 0x1000 + data_offset
            value = self.read(data_len)
        else:
            name = self.read(key_len)
            value = self.read(data_len)
        return name, value, next_offset

    def read_all(self) -> dict:
        result = {}

        page_offset = 2 * self.page_size
        while page_offset < self.size:
            self.pos = page_offset
            next_offset = self.read_int(2)
            while next_offset:
                self.pos = page_offset + next_offset
                k, v, next_offset = self.read_cell()
                result[k.decode()] = v.decode()
            page_offset += self.page_size

        return result



import asyncio
import aiohttp
import functools
import concurrent.futures
import logging
import threading
import urllib.request
from urllib.parse import urlparse
from urllib.parse import parse_qs
from .. import utils
from .. import fhem


# PyChromecast
import pychromecast
from pychromecast.error import ChromecastConnectionError

# YouTube
from pychromecast.controllers.youtube import YouTubeController

# DashCast
import pychromecast.controllers.dashcast as dashcast

# Spotify
from pychromecast.controllers.spotify import SpotifyController
import spotipy
from spotipy.oauth2 import SpotifyClientCredentials

# youtube_dl
import youtube_dl

connection_update_lock = threading.Lock()


class googlecast:
    def __init__(self, logger):
        self.logger = logger
        # do only error logging for pychromecast library
        logging.getLogger("pychromecast").setLevel(logging.ERROR)
        self.cast = None
        self.loop = asyncio.get_running_loop()
        self.hash = None
        self.currPosTask = None
        self.connectionStateCache = ""
        self.browser = None

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        if len(args) > 3:
            hash["CASTNAME"] = args[3]
        self.hash = hash

        # set userattr as long as Define can't be called on startup
        await fhem.setDevAttrList(
            hash["NAME"], "favorite_1 favorite_2 favorite_3 favorite_4 favorite_5"
        )

        if self.browser:
            pychromecast.stop_discovery(self.browser)
        if self.cast:
            self.cast.disconnect()

        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdateIfChanged(hash, "state", "offline")
        await fhem.readingsBulkUpdateIfChanged(hash, "connection", "disconnected")
        await fhem.readingsEndUpdate(hash, 1)

        self.startDiscovery()

        return ""

    # FHEM FUNCTION
    async def Undefine(self, hash):
        try:
            pychromecast.stop_discovery(self.browser)
            if self.cast:
                self.cast.disconnect()
        except:
            pass

    # FHEM FUNCTION
    async def Set(self, hash, args, argsh):
        if len(args) < 2 or args[1] == "?":
            return (
                "Unknown argument ?, choose one of "
                "stop:noArg pause:noArg rewind:noArg skip:noArg quitApp:noArg "
                "play addToQueue playFavorite:1,2,3,4,5 volume:slider,0,1,100 seek "
                "next:noArg prev:noArg subtitles:on,off "
                "displayWebsite speak startApp "
                "volUp:noArg volDown:noArg"
            )
        else:
            if self.connectionStateCache != "CONNECTED":
                return "Please wait until connected..."

            action = args[1]
            if action == "play":
                url = ""
                if "url" in argsh:
                    url = argsh["url"]
                elif len(args) > 2:
                    url = args[2]

                if len(url) > 0:
                    self.playUrl(url)
                else:
                    self.cast.media_controller.play()
            elif action == "addToQueue":
                url = ""
                if "url" in argsh:
                    url = argsh["url"]
                elif len(args) > 2:
                    url = args[2]

                if len(url) > 0:
                    self.queueUrl(url)
            elif action == "playFavorite":
                url = await fhem.AttrVal(self.hash["NAME"], "favorite_" + args[2], "")
                if len(url) == 0:
                    return "Please set favorite before usage"
                self.playUrl(url)
            elif action == "stop":
                self.cast.media_controller.stop()
            elif action == "pause":
                self.cast.media_controller.pause()
            elif action == "quitApp":
                self.cast.quit_app()
            elif action == "startApp":
                if len(args) > 2:
                    appId = args[2]
                    self.cast.start_app(appId)
                else:
                    return "Please specify app_id as parameter."
            elif action == "skip":
                self.cast.media_controller.skip()
            elif action == "rewind":
                self.cast.media_controller.rewind()
            elif action == "seek":
                position = 0
                if len(args) > 2:
                    position = args[2]
                self.cast.media_controller.seek(position)
            elif action == "next":
                self.cast.media_controller.queue_next()
            elif action == "prev":
                self.cast.media_controller.queue_prev()
            elif action == "volUp":
                self.cast.volume_up()
            elif action == "volDown":
                self.cast.volume_down()
            elif action == "subtitles":
                onoff = "on"
                if len(args) > 3:
                    onoff = args[2]
                if onoff == "on":
                    self.cast.media_controller.enable_subtitle(1)
                else:
                    self.cast.media_controller.disable_subtitle()
            elif action == "volume":
                vol = int(args[2])
                self.cast.set_volume(vol / 100)
            elif action == "speak":
                txt = args[2]
                ttsUrl = (
                    "http://translate.google.com/translate_tts?tl=de&client=tw-ob&q="
                    + urllib.parse.quote(txt)
                )
                self.cast.play_media(ttsUrl, "audio/mpeg")
            elif action == "displayWebsite":
                dashUrl = "https://fhem.de/"
                if "url" in argsh:
                    dashUrl = argsh["url"]
                elif len(args) > 2:
                    dashUrl = args[2]
                self.loop.create_task(self.displayWebsite(dashUrl))

    def playUrl(self, url):
        videoid = self.extract_video_id(url)
        if videoid == None:
            if url.find("spotify") >= 0:
                self.loop.create_task(self.playSpotifyThread(url))
            else:
                self.loop.create_task(self.playDefaultMedia(url))
        else:
            if self.cast.cast_type == "audio":
                self.loop.create_task(self.playYoutubeAudio(url))
            else:
                playlistid = self.extract_playlist_id(url)
                self.loop.create_task(self.playYoutube(videoid, playlistid))

    def queueUrl(self, url):
        self.loop.create_task(self.playDefaultMedia(url, enqueue=True))

    async def playDefaultMedia(self, uri, enqueue=False):
        try:
            session = aiohttp.ClientSession()
            res = await session.get(uri)
            mime = res.headers["Content-Type"]
            self.cast.play_media(uri, mime, enqueue=enqueue)
        except:
            self.logger.exception(f"Failed to play: {uri}")

    async def playSpotifyThread(self, uri):
        with concurrent.futures.ThreadPoolExecutor() as pool:
            await self.loop.run_in_executor(
                pool, functools.partial(self.playSpotify, uri)
            )

    async def playSpotify(self, uri):
        # FIXME user needs to enter CLIENT_ID and CLIENT_SECRET from Spotify Dashboard
        client_credentials_manager = SpotifyClientCredentials(
            "CLIENT_ID", "CLIENT_SECRET"
        )
        data = client_credentials_manager.get_access_token()
        access_token = data["access_token"]
        expires = data["expires_in"]

        # Create a spotify client
        client = spotipy.Spotify(auth=access_token)

        # Launch the spotify app on the cast we want to cast to
        sp = SpotifyController(access_token, expires)
        self.cast.register_handler(sp)
        sp.launch_app()

        if not sp.is_launched and not sp.credential_error:
            self.logger.error("Failed to launch spotify controller due to timeout")
            return
        if not sp.is_launched and sp.credential_error:
            self.logger.error(
                "Failed to launch spotify controller due to credential error"
            )
            return

        # Query spotify for active devices
        devices_available = client.devices()

        # Match active spotify devices with the spotify controller's device id
        for device in devices_available["devices"]:
            if device["id"] == sp.device:
                spotify_device_id = device["id"]
                break

        if not spotify_device_id:
            self.logger.error(
                'No device with id "{}" known by Spotify'.format(sp.device)
            )
            self.logger.error("Known devices: {}".format(devices_available["devices"]))
            return

        # Start playback
        if uri.find("track") > 0:
            client.start_playback(device_id=spotify_device_id, uris=uri)
        else:
            client.start_playback(device_id=spotify_device_id, context_uri=uri[0])

    async def displayWebsite(self, url):
        d = dashcast.DashCastController()
        self.cast.register_handler(d)
        d.load_url(url, force=True, reload_seconds=30)

    async def playYoutubeAudio(self, uri):
        with concurrent.futures.ThreadPoolExecutor() as pool:
            video_url = await self.loop.run_in_executor(
                pool, functools.partial(self.getYoutubeAudioUrl, uri)
            )
            self.cast.play_media(video_url, "audio/mp4")

    def getYoutubeAudioUrl(self, uri):
        ydl = youtube_dl.YoutubeDL(
            {
                "forceurl": True,
                "simulate": True,
                "quiet": "1",
                "no_warnings": "1",
                "skip_download": True,
                "format": "bestaudio/best",
                "youtube_include_dash_manifest": True,
            }
        )
        result = ydl.extract_info(uri, download=False)
        if "entries" in result:
            # Can be a playlist or a list of videos
            video = result["entries"][0]
        else:
            # Just a video
            video = result
        video_url = video["url"]
        return video_url

    async def playYoutube(self, videoid, playlistid):
        yt = YouTubeController()
        self.cast.register_handler(yt)
        await utils.run_blocking(functools.partial(yt.play_video, videoid, playlistid))

    def extract_video_id(self, url):
        # Examples:
        # - http://youtu.be/SA2iWivDJiE
        # - http://www.youtube.com/watch?v=_oPAwA_Udwc&feature=feedu
        # - http://www.youtube.com/embed/SA2iWivDJiE
        # - http://www.youtube.com/v/SA2iWivDJiE?version=3&amp;hl=en_US
        query = urlparse(url)
        if query.hostname == "youtu.be":
            return query.path[1:]
        if query.hostname in ("www.youtube.com", "youtube.com"):
            if query.path == "/watch":
                return parse_qs(query.query)["v"][0]
            if query.path[:7] == "/embed/":
                return query.path.split("/")[2]
            if query.path[:3] == "/v/":
                return query.path.split("/")[2]
        # fail?
        return None

    def extract_playlist_id(self, url):
        query = urlparse(url)
        if "list" in parse_qs(query.query):
            return parse_qs(query.query)["list"]
        return None

    def startDiscovery(self):
        def castFound(chromecast):
            if chromecast.name == self.hash["CASTNAME"] and self.cast is None:
                self.logger.info("Discovered cast: " + chromecast.name)
                self.cast = chromecast
                # add status listener
                self.cast.register_connection_listener(self)
                self.cast.register_status_listener(self)
                # add media controller listener
                self.cast.media_controller.register_status_listener(self)
                self.logger.debug("wait for chromecast")
                # timeout 0.001 just waits for status to be ready
                # but we just need the thread to start by calling wait()
                self.cast.wait(0.001)
                self.logger.debug("wait finished")

        self.logger.debug("Start discovery")
        self.browser = pychromecast.get_chromecasts(
            blocking=False, tries=None, retry_wait=5, timeout=5, callback=castFound
        )

    # THREADING: this function is called by run_once pychromecast thread
    def new_connection_status(self, status):
        # connection update might come from different threads
        connection_update_lock.acquire()
        self.logger.debug("new_connection_status: " + status.status)

        # prevent to many disconnect events
        if status.status != self.connectionStateCache:
            self.connectionStateCache = status.status
            # run reading updates in main thread
            res = asyncio.run_coroutine_threadsafe(
                self.updateConnectionReadings(self.hash, status), self.loop
            )
            res.result()

        if status.status == "CONNECTED":
            # run reading updates in main thread
            res = asyncio.run_coroutine_threadsafe(
                self.updateDeviceReadings(self.hash), self.loop
            )
            res.result()
        connection_update_lock.release()

    # THREADING: this function is called by run_once pychromecast thread
    def new_cast_status(self, status):
        self.logger.debug("new_cast_status")
        # run reading updates in main thread
        res = asyncio.run_coroutine_threadsafe(
            self.updateStatusReadings(self.hash, status), self.loop
        )
        res.result()

    # THREADING: this function is called by run_once pychromecast thread
    def new_media_status(self, status):
        if (
            status.player_state == "PLAYING"
            and self.currPosTask == None
            and status.duration
            and status.duration > 0
        ):
            self.currPosTask = self.loop.create_task(self.updateCurrentPosition())
        elif status.player_state != "PLAYING":
            if self.currPosTask:
                self.currPosTask.cancel()
                self.currPosTask = None
        self.logger.debug("new_media_status")
        # run reading updates in main thread
        res = asyncio.run_coroutine_threadsafe(
            self.updateMediaStatusReadings(self.hash, status), self.loop
        )
        res.result()

    async def updateCurrentPosition(self):
        try:
            while True:
                await asyncio.sleep(30)
                self.cast.media_controller.update_status()
        except concurrent.futures.CancelledError:
            # task was cancelled, nothing playing
            pass
        except Exception:
            self.logger.error("Update media status failed", exc_info=True)

    async def updateConnectionReadings(self, hash, status):
        self.logger.debug("updateConnectionReading")
        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdateIfChanged(
            hash, "connection", status.status.lower()
        )
        if status.status == "CONNECTED":
            await fhem.readingsBulkUpdateIfChanged(hash, "state", "online")
        else:
            await fhem.readingsBulkUpdateIfChanged(hash, "state", "offline")
        await fhem.readingsEndUpdate(hash, 1)

    async def updateStatusReadings(self, hash, status):
        self.logger.debug("updateStatusReadings")
        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdateIfChanged(
            hash, "volume", round(status.volume_level * 100)
        )
        await fhem.readingsBulkUpdateIfChanged(
            hash, "is_active_input", status.is_active_input
        )
        await fhem.readingsBulkUpdateIfChanged(hash, "is_stand_by", status.is_stand_by)
        await fhem.readingsBulkUpdateIfChanged(hash, "mute", status.volume_muted)
        await fhem.readingsBulkUpdateIfChanged(
            hash, "display_name", status.display_name
        )
        # await fhem.readingsBulkUpdateIfChanged(hash, "namespaces", status.namespaces)
        await fhem.readingsBulkUpdateIfChanged(hash, "session_id", status.session_id)
        await fhem.readingsBulkUpdateIfChanged(
            hash, "transport_id", status.transport_id
        )
        await fhem.readingsBulkUpdateIfChanged(hash, "status_text", status.status_text)
        await fhem.readingsBulkUpdateIfChanged(hash, "icon_url", status.icon_url)
        await fhem.readingsBulkUpdateIfChanged(hash, "app_id", status.app_id)
        await fhem.readingsEndUpdate(hash, 1)

    async def updateMediaStatusReadings(self, hash, status):
        self.logger.debug("updateMediaStatusReadings")
        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdateIfChanged(
            hash, "mediaPlayerState", status.player_state
        )
        await fhem.readingsBulkUpdateIfChanged(
            hash, "mediaContentId", status.content_id
        )
        await fhem.readingsBulkUpdateIfChanged(
            hash, "mediaContentType", status.content_type
        )
        await fhem.readingsBulkUpdateIfChanged(hash, "mediaDuration", status.duration)
        if status.current_time:
            await fhem.readingsBulkUpdateIfChanged(
                hash, "mediaCurrentPosition", round(status.current_time)
            )
        else:
            await fhem.readingsBulkUpdateIfChanged(hash, "mediaCurrentPosition", "")
        if status.duration and status.current_time:
            await fhem.readingsBulkUpdateIfChanged(
                hash,
                "mediaCurrentPosPercent",
                round(status.current_time / status.duration * 100),
            )
        else:
            await fhem.readingsBulkUpdateIfChanged(hash, "mediaCurrentPosPercent", "")
        await fhem.readingsBulkUpdateIfChanged(
            hash, "mediaStreamType", status.stream_type
        )
        await fhem.readingsBulkUpdateIfChanged(hash, "mediaTitle", status.title)
        await fhem.readingsBulkUpdateIfChanged(
            hash, "mediaSeriesTitle", status.series_title
        )
        await fhem.readingsBulkUpdateIfChanged(hash, "mediaSeason", status.season)
        await fhem.readingsBulkUpdateIfChanged(hash, "mediaEpisode", status.episode)
        await fhem.readingsBulkUpdateIfChanged(hash, "mediaArtist", status.artist)
        await fhem.readingsBulkUpdateIfChanged(hash, "mediaAlbum", status.album_name)
        await fhem.readingsBulkUpdateIfChanged(
            hash, "mediaAlbumArtist", status.album_artist
        )
        await fhem.readingsBulkUpdateIfChanged(hash, "mediaTrack", status.track)
        if len(status.images) > 0:
            await fhem.readingsBulkUpdateIfChanged(
                hash, "mediaImageUrl", status.images[0].url
            )
            await fhem.readingsBulkUpdateIfChanged(
                hash, "mediaImageHeight", status.images[0].height
            )
            await fhem.readingsBulkUpdateIfChanged(
                hash, "mediaImageWidth", status.images[0].width
            )
        else:
            await fhem.readingsBulkUpdateIfChanged(hash, "mediaImageUrl", "")
            await fhem.readingsBulkUpdateIfChanged(hash, "mediaImageHeight", "")
            await fhem.readingsBulkUpdateIfChanged(hash, "mediaImageWidth", "")

        if status.player_state == "PLAYING":
            await fhem.readingsBulkUpdateIfChanged(hash, "state", "playing")
        elif status.player_state == "BUFFERING":
            await fhem.readingsBulkUpdateIfChanged(hash, "state", "buffering")
        elif status.player_state == "PAUSED":
            await fhem.readingsBulkUpdateIfChanged(hash, "state", "paused")
        else:
            if (
                await fhem.ReadingsVal(hash["NAME"], "connection", "failed")
                == "connected"
            ):
                await fhem.readingsBulkUpdateIfChanged(hash, "state", "online")
            else:
                await fhem.readingsBulkUpdateIfChanged(hash, "state", "offline")
        await fhem.readingsEndUpdate(hash, 1)

    async def updateDeviceReadings(self, hash):
        self.logger.debug("updateDeviceReadings")
        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdateIfChanged(hash, "name", self.cast.name)
        await fhem.readingsBulkUpdateIfChanged(hash, "model_name", self.cast.model_name)
        await fhem.readingsBulkUpdateIfChanged(hash, "uuid", self.cast.uuid)
        await fhem.readingsBulkUpdateIfChanged(hash, "cast_type", self.cast.cast_type)
        # await fhem.readingsBulkUpdateIfChanged(hash, "uri", self.cast.uri)
        await fhem.readingsBulkUpdateIfChanged(hash, "ignore_cec", self.cast.ignore_cec)
        await fhem.readingsEndUpdate(hash, 1)

        if await fhem.AttrVal(hash["NAME"], "icon", "") == "":
            if await fhem.ReadingsVal(hash["NAME"], "cast_type", "cast") == "cast":
                await fhem.CommandAttr(
                    self.hash, self.hash["NAME"] + " icon scene_scene"
                )
            else:
                await fhem.CommandAttr(
                    self.hash, self.hash["NAME"] + " icon gassistant"
                )
            await fhem.CommandAttr(hash, "cmdIcon pause:rc_PAUSE play:rc_PLAY")
            await fhem.CommandAttr(hash, "webCmd volume:play:pause")



import asyncio
import codecs
import functools
import binascii
import time
import struct

from bluepy import btle

from .. import fhem
from .. import utils

DEFAULT_TIMEOUT = 1

HANDLE_RW_PASSWORD = 0x0048
HANDLE_RW_WATERING = 0x0015
HANDLE_R_BATTERY = 0x0039
HANDLE_R_TEMPERATURE = 0x003B
HANDLE_R_MIN_TEMP = 0x003D
HANDLE_R_MAX_TEMP = 0x003F
HANDLE_R_FIRMWARE = 0x004E
HANDLE_RW_DEVNAME = 0x0052
HANDLE_RW_ECO_PART1 = 0x0033
HANDLE_RW_ECO_PART2 = 0x0045
HANDLE_RW_TIME_OFFSET = 0x0035
HANDLE_R_MAC = 0x004A
HANDLE_RW_INCREASEREDUCE = 0x0043
HANDLE_W_COMMITCODE = 0x0037
HANDLE_RW_TIMERS = [
    "0x0017",
    "0x0019",
    "0x001b",
    "0x001d",
    "0x001f",
    "0x0021",
    "0x0023",
    "0x0031",
    "0x0025",
    "0x0027",
    "0x0029",
    "0x002b",
    "0x002d",
    "0x002f",
]


class gfprobt:
    def __init__(self, logger):
        self.logger = logger
        self._conn = btle.Peripheral()
        self._conn.withDelegate(self)
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        self.hash = hash
        self._mac = args[3]
        self.hash["MAC"] = self._mac
        asyncio.create_task(self.update())
        return "Module not ready yet"

    # FHEM FUNCTION
    async def Undefine(self, hash):
        return

    # FHEM FUNCTION
    async def Set(self, hash, args, argsh):
        return

    async def update(self):
        await utils.run_blocking(functools.partial(self.blocking_update))
        await fhem.readingsBeginUpdate(self.hash)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "state", self._watering)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "watering", self._watering)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "battery", self._battery)
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "temperature", self._temperature
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "min_temperature", self._min_temp
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "max_temperature", self._max_temp
        )
        await fhem.readingsBulkUpdateIfChanged(self.hash, "firmware", self._firmware)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "devname", self._devname)
        await fhem.readingsEndUpdate(self.hash, 1)

    def blocking_connect(self):
        self._conn.connect()

    def blocking_update(self):
        self._conn.connect(self._mac)
        self._conn.writeCharacteristic(HANDLE_RW_PASSWORD, b"123456")
        self._watering = struct.unpack(
            "<b", self._conn.readCharacteristic(HANDLE_RW_WATERING)
        )[0]
        self._battery = struct.unpack(
            "<h", self._conn.readCharacteristic(HANDLE_R_BATTERY)
        )[0]
        self._temperature = struct.unpack(
            "<h", self._conn.readCharacteristic(HANDLE_R_TEMPERATURE)
        )[0]
        self._min_temp = struct.unpack(
            "<h", self._conn.readCharacteristic(HANDLE_R_MIN_TEMP)
        )[0]
        self._max_temp = struct.unpack(
            "<h", self._conn.readCharacteristic(HANDLE_R_MAX_TEMP)
        )[0]
        self._firmware = self._conn.readCharacteristic(HANDLE_R_FIRMWARE)
        self._firmware = self._firmware[1] + "." + self._firmware[0]
        self._devname = self._conn.readCharacteristic(HANDLE_RW_DEVNAME).decode("utf-8")
        self._eco = self._conn.readCharacteristic(HANDLE_RW_ECO_PART1)
        self._eco += " " + self._conn.readCharacteristic(HANDLE_RW_ECO_PART2)
        self._timeoffset = struct.unpack(
            "<I", self._conn.readCharacteristic(HANDLE_RW_TIME_OFFSET)
        )[0]
        self._devmac = self._conn.readCharacteristic(HANDLE_R_MAC)
        self._increasereduce = self._conn.readCharacteristic(HANDLE_RW_INCREASEREDUCE)
        self._raw_timers = {}
        for handle_timer in HANDLE_RW_TIMERS:
            self._raw_timers[handle_timer] = self._conn.readCharacteristic(handle_timer)
            self.logger.debug(self._raw_timers[handle_timer])
        self._conn.disconnect()

    def write_offset(self):
        now = time.localtime()
        sec_since_mon = (
            now.tm_wday * 3600 * 24 + now.tm_hour * 3600 + now.tm_min * 60 + now.tm_sec
        )
        sec_since_mon = struct.pack("<I", sec_since_mon)
        self._conn.writeCharacteristic(HANDLE_RW_TIME_OFFSET, sec_since_mon)

    def commit_code(self):
        self._conn.writeCharacteristic(HANDLE_W_COMMITCODE, b"00")
        self._conn.writeCharacteristic(HANDLE_W_COMMITCODE, b"01")



import asyncio
import logging
import bluetooth
import functools
from bt_proximity import BluetoothRSSI

from .. import fhem
from .. import utils


class bt_presence:
    def __init__(self, logger):
        self.logger = logger
        self.hash = None
        self.btscan_task = None
        self._interval = {"absent": 10, "present": 60}
        self._threshold = {"absent": 0, "present": 0}
        self._curr_state = ""
        self._count_diff_state = 0

    def lookup_name(self, mac):
        return bluetooth.lookup_name(mac, timeout=5)

    async def run_bt_scan(self):
        curr_name = ""
        curr_rssi = 0
        while True:
            new_state = "absent"
            try:
                # check max 3 times for device_name
                for i in range(0, 2):
                    device_name = await utils.run_blocking(
                        functools.partial(self.lookup_name, self._address)
                    )
                    if device_name:
                        break

                if device_name:
                    self._btrssi = BluetoothRSSI(self._address)
                    rssi = await utils.run_blocking(
                        functools.partial(self._btrssi.request_rssi)
                    )
                    self._btrssi.close()
                    rssi = rssi[0]
                    if curr_name != device_name:
                        curr_name = device_name
                        await fhem.readingsSingleUpdateIfChanged(
                            self.hash, "name", device_name, 1
                        )
                    if curr_rssi != rssi:
                        curr_rssi = rssi
                        await fhem.readingsSingleUpdateIfChanged(
                            self.hash, "rssi", rssi, 1
                        )
                    new_state = "present"
            except:
                self.logger.exception("BleakScanner failed")

            if self._curr_state != new_state:
                self._count_diff_state += 1
                if self._count_diff_state >= self._threshold[new_state]:
                    self._curr_state = new_state
                    self._count_diff_state = 0
                    await self.update_state(self._curr_state)

            await asyncio.sleep(self._interval[self._curr_state])

    async def update_state(self, new_state):
        await fhem.readingsSingleUpdateIfChanged(self.hash, "presence", new_state, 1)
        await fhem.readingsSingleUpdateIfChanged(self.hash, "state", new_state, 1)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        self.hash = hash
        if len(args) < 4:
            return "Usage: define p_mysmartphone PythonModule bt_presence <MAC>"

        self._address = args[3]
        self.hash["MAC"] = args[3]

        await fhem.addToDevAttrList(
            hash["NAME"],
            "absentInterval presentInterval absentThreshold presentThreshold",
        )

        self._interval = {
            "absent": int(await fhem.AttrVal(hash["NAME"], "absentInterval", "60")),
            "present": int(await fhem.AttrVal(hash["NAME"], "presentInterval", "60")),
        }
        self._threshold = {
            "absent": int(await fhem.AttrVal(hash["NAME"], "absentThreshold", "0")),
            "present": int(await fhem.AttrVal(hash["NAME"], "presentThreshold", "0")),
        }

        if self.btscan_task:
            self.btscan_task.cancel()
        self.btscan_task = asyncio.create_task(self.run_bt_scan())
        return ""

    # FHEM FUNCTION
    async def Undefine(self, hash):
        if self.btscan_task:
            self.btscan_task.cancel()

    async def Attr(self, hash, args, argsh):
        cmd = args[0]
        name = args[1]
        attr_name = args[2]
        attr_val = args[3]

        if cmd == "set":
            if attr_name == "absentInterval":
                self._interval["absent"] = int(attr_val)
            elif attr_name == "presentInterval":
                self._interval["present"] = int(attr_val)
            elif attr_name == "absentThreshold":
                self._threshold["absent"] = int(attr_val)
            elif attr_name == "presentThreshold":
                self._threshold["present"] = int(attr_val)
        else:
            if attr_name == "absentInterval":
                self._interval["absent"] = 10
            elif attr_name == "presentInterval":
                self._interval["present"] = 60
            elif attr_name == "absentThreshold":
                self._threshold["absent"] = 0
            elif attr_name == "presentThreshold":
                self._threshold["present"] = 0
